#  Creating starting from tempo.g 
#/*
# * Tempo grammar: for dparser
# *
# * Based on Appendix C of the Tempo User Guide
# *
# * This should be enough to parse all the examples in Tempo/examples/newUserGuide
# * (except twoTaskRace.tioa because it includes a simulation relation)
# */
#
#/*
# * Top-level:
# * - removed 'simDef and simProg'
# * - simplified the invariant syntax
# */
from dparser import Parser, SyntaxError
import sys
import os

def d_spec(s, nodes):
    "spec : ( typeDecls | imports | funDecls | vocabDef | autoDef | invDef )+ "

def d_imports(s, nodes):
    "imports : 'imports' vocabRef (',' vocabRef )* "

def d_vocabRef(s, nodes): 
    "vocabRef : ID ( '(' actual (',' actual )* ')' )? "

def d_funDecls(s, nodes): 
    "funDecls : 'let' (funDecl ';')+ "

def d_funDecl(s, nodes): 
    "funDecl : ID '(' (ID (',' ID)* )? ')' ':' typeSignature '=' expr "

def d_invDef(s, nodes): 
    "invDef : 'invariant' idOrNumeral? 'of' ID ':' expr ';' "

def d_idOrNumerals(s, nodes): 
    "idOrNumerals : idOrNumeral ( ',' idOrNumeral )* "

def d_idOrNumeral(s, nodes): 
    "idOrNumeral : (ID | INT ) "

def d_actual(s, nodes): 
    "actual : (expr | 'Type' typeRef ) "


# /*
#  * Type declarations
#  */
def d_typeDecls(s, nodes): 
    "typeDecls : 'types' typeDecl ( ',' typeDecl )* "

def d_typeDecl(s, nodes): 
    "typeDecl : ID (':' typeRef)? "

def d_typeRef(s, nodes): 
    "typeRef : ID | ID '[' typeList ']' | 'Enumeration' '[' idOrNumerals ']' | 'Tuple' '[' fieldDecls ']' | 'Union' '[' fieldDecls ']' "

def d_fieldDecls(s, nodes): 
    "fieldDecls : fieldDecl ( ',' fieldDecl )* "

def d_fieldDecl(s, nodes): 
    "fieldDecl : ID ( ',' ID )* ':' typeRef "

def d_typeSignature(s, nodes): 
    "typeSignature : (typeList)? '->' typeRef "

def d_typeList(s, nodes): 
    "typeList : typeRef ( ',' typeRef )* "


#/*
# * Vocabulary definition: we don't support the infix/mixfix/prefix specs
# * operator declarations must be of the form ID: typeSignature
# */
def d_vocabDef(s, nodes): 
    "vocabDef : 'vocabulary' ID (formals)? (importsAndTypes)? (operators)? 'end' "

def d_importsAndTypes(s, nodes): 
    "importsAndTypes : ( imports | typeDecls )+ "

def d_operators(s, nodes): 
    "operators : 'operators' opDecl (',' opDecl )* "

def d_opDecl(s, nodes): 
    "opDecl : ID ( ',' ID )* ':' typeSignature "



#/*
# * Automaton definitions
# * - no support for tasks and schedules
# */
def d_autoDef(s, nodes): 
    "autoDef : 'automaton' ID ( formals (where)? )? (imports)? autoCore "

def d_formals(s, nodes): 
    "formals : '(' formal (',' formal )* ')' "

def d_formal(s, nodes): 
    "formal : ID (',' ID)* ':' (typeRef | 'Type') "

def d_where(s, nodes): 
    "where : 'where' expr "


def d_autoCore(s, nodes): 
    "autoCore : basicAutomaton | composedAutomaton "

def d_basicAutomaton(s, nodes): 
    "basicAutomaton : (actionSignature)? states (funDecls)? (transitions)? (trajectories)? "


#/*
# * Signature
# */
def d_actionSignature(s, nodes): 
    "actionSignature : 'signature' (formalActions)+ "

def d_formalActions(s, nodes): 
    "formalActions : \
      'input' formalAction ( ',' formalAction )* \
    | 'output' formalAction ( ',' formalAction )* \
    | 'internal' formalAction ( ',' formalAction )* "

def d_formalAction(s, nodes): 
    "formalAction : ID ( '(' sigFormal (',' sigFormal )* ')' (where)? )? "

def d_sigFormal(s, nodes): 
    "sigFormal : 'const' expr | ID (',' ID )* ':' typeRef "



#/*
# * State variable declaration + initialization
## * - note: initially seems to be an old form (could remove it?)
# */
def d_states(s, nodes): 
    "states : 'states' (state ';')+ (initially)? | 'states' "

def d_state(s, nodes): 
    "state : ID ':' typeRef ( ':=' value )? "

def d_initially(s, nodes): 
    "initially : 'initially' expr ';' "




#/*
# * Transitions: allowed only 'var := value' and if-then-else in the 'effStmt'
# * Also removed the 'ensuring' clause and 'urgency'
# *      simplified the precondition syntax to 'pre' expr ';'
# */
def d_transitions(s, nodes): 
    "transitions : 'transitions' (transition)+ "

def d_transition(s, nodes): 
    "transition : \
      'input' transitionCore  \
    | 'output' transitionCore  \
    | 'internal' transitionCore "

def d_transitionCore(s, nodes): 
    "transitionCore : ID ( '(' actionActuals ')' (where)? )? (localVars)? (precondition)? (effect)? "

def d_actionActuals(s, nodes): 
    "actionActuals : expr (',' expr)* "

def d_localVars(s, nodes): 
    "localVars : 'locals' ( localDecl ';' )+ "

def d_precondition(s, nodes): 
    "precondition : 'pre' expr ';' "

def d_effect(s, nodes): 
    "effect : 'eff' effProgram  "

def d_localDecl(s, nodes): 
    "localDecl : ID ':' typeRef ( ':=' value ) ? "

def d_effProgram(s, nodes): 
    "effProgram : ( effStmt )+ "

def d_effStmt(s, nodes): 
    "effStmt : lvalue ':=' value \
    | 'if' effCondRec 'fi' \
    | ';' "

def d_effCondRec(s, nodes): 
    "effCondRec : expr 'then' effProgram ( 'elseif' effCondRec | 'else' effProgram ) ? "



#/*
# * Trajectories:
# * - simplified the syntax: 
# *   Formals and local functions are not supported. 
# *   They don't seem to be use in any of the examples anyway.
# * - invariant clause: simplified to 'invariant expr ;'
# */
def d_trajectories(s, nodes): 
    "trajectories : 'trajectories' (trajectory)+ "

def d_trajectory(s, nodes): 
    "trajectory : 'trajdef' ID (trajInvariant)? (stopCond)? (evolve)? "

def d_trajInvariant(s, nodes): 
    "trajInvariant : 'invariant' expr ';' "

def d_stopCond(s, nodes): 
    "stopCond : 'stop' 'when' expr ';' "

def d_evolve(s, nodes): 
    "evolve : 'evolve' ( evolveStmt)+  "

def d_evolveStmt(s, nodes): 
    "evolveStmt : expr ';' "



#/*
# * Composed Automaton
# */
def d_composedAutomaton(s, nodes): 
    "composedAutomaton : components (hiddenActionSets )? "

def d_components(s, nodes): 
    "components : 'components' ( component ';' )+ "

def d_component(s, nodes): 
    "component : ID ('[' varList (',' varList )* ']' )? (':' componentDef )? (where)? "

def d_componentDef(s, nodes): 
    "componentDef : ID ( '(' actual (',' actual )* ')' )? "

def d_hiddenActionSets(s, nodes): 
    "hiddenActionSets : 'hidden' ( actionSet ';' ) + "

def d_actionSet(s, nodes): 
    "actionSet : ( compInstance '.')? ID ('(' expr (',' expr )* ')' )? "

def d_compInstance(s, nodes): 
    "compInstance : ID ('[' expr (',' expr )* ']' )? ( '.' ( '[' expr (',' expr)* ']' )? )* "

def d_varLis(s, nodes):
    "varList: ID (',' ID)* ':' typeRef "


#/*
# * Expressions
# */
def d_expr(s, nodes): 
    "expr : \
      'if' expr 'then' expr 'else' expr \
    | expr BINARY_OP expr \
    | UNARY_OP expr \
    | QUANTIFIER ID ':' typeRef expr \
    | '[' expr (',' expr)* ']' \
    | '{' ID ':' typeRef 'where' expr '}' \
    | '{' expr (',' expr)* '}' \
    | expr ( '.' ID | '[' expr (',' expr)* ']' )+ \
    | ID '(' expr (',' expr)* ')' \
    | '(' typeRef ')' expr \
    | 'true' \
    | 'false' \
    | '\infty' \
    | '{}' \
    | INT \
    | FLOAT \
    | ID \
    | ID '\\'' \
    | '(' expr ')' "



#/*
# * Assignment
# */
def d_lvalue(s, nodes): 
    "lvalue : ID ( '[' expr (',' expr)* ']' | '.' ID )* "

def d_value(s, nodes): 
    "value : expr | choice "

def d_choice(s, nodes): 
    "choice : 'choose' variable 'where' expr "

def d_variable(s, nodes): 
    "variable : ID (':' typeRef)? "


#/*
# * TERMINALS
# */
def d_ID(s, nodes): 
    'ID : "[a-zA-Z][a-zA-Z_0-9\$]*" '

def d_INT(s, nodes): 
    'INT : "[\+\-]?[0-9]+" '

def d_FLOAT(s, nodes): 
    'FLOAT : "[\+\-]?[0-9]+\.[0-9]+" \
  | "[\+\-]?[0-9]+(\.[0-9]+)? [Ee][\+\-][0-9]+" '

def d_QUANTIFIER(s, nodes): 
    "QUANTIFIER : '\A' | '\E' "

def d_UNARY_OP(s, nodes): 
    "UNARY_OP : \
   '+'   $unary_op_right 40 | \
   '-'   $unary_op_right 40 | \
   '~'   $unary_op_right 40 "

def d_BINARY_OP(s, nodes): 
    "BINARY_OP : \
   '**'        $binary_op_left 30 | \
   '/'         $binary_op_left 29 | \
   '*'         $binary_op_left 29 | \
   '+'         $binary_op_left 28 | \
   '-'         $binary_op_left 28 | \
   '\I'        $binary_op_left 27 | \
   '\U'        $binary_op_left 26 | \
   '|-'        $binary_op_left 25 | \
   '-|'        $binary_op_left 25 | \
   '||'        $binary_op_left 24 | \
   '<'         $binary_op_left 20 | \
   '>'         $binary_op_left 20 | \
   '<='        $binary_op_left 20 | \
   '>='        $binary_op_left 20 | \
   '\notin'    $binary_op_left 20 | \
   '\in'       $binary_op_left 20 | \
   '\subset'   $binary_op_left 20 | \
   '\subseteq' $binary_op_left 20 | \
   '\supset'   $binary_op_left 20 | \
   '\supseteq' $binary_op_left 20 | \
   '='         $binary_op_left 19 | \
   '~='        $binary_op_left 19 | \
   '/\\x5c'     $binary_op_left 6 | \
   '\/'        $binary_op_left 5 | \
   '=>'        $binary_op_left 4 | \
   '<=>'       $binary_op_left 3 "


def d_whitespace(s, nodes): 
    'whitespace : ( "[ \t\n\r\v\f]" | comment )* '

def d_comment(s, nodes): 
    'comment :  "%" "[^\n]*" "\n"  '

def printUsage():
    print """
NAME: dtempo - termpo parser 
SYNOPSIS: dtempo <filename.tioa>
"""
    sys.exit()

def main_aux(filename):
    with open(filename, 'r') as f:
        x = f.read()
    try:
        y = Parser().parse(x)    # y is now a dparser.ParsedStructure instance
    except SyntaxError as e:
        print e
        return filename
    z = y.getStructure()
    #print z
    #print z.toprettyxml()
    return filename

def main():
    args = sys.argv[1:]
    if len(args) < 1 or len(args) > 1 or args[0][0] == '-':
        printUsage()
        return 1
    filename = args[0]
    if not(os.path.isfile(filename)):
        print "File does not exist. Quitting."
        return 1
    output_filename = main_aux(filename)
    return output_filename


if __name__ == '__main__':
    main()
