/*
 * Tempo grammar: for dparser
 *
 * Based on Appendix C of the Tempo User Guide
 *
 * This should be enough to parse all the examples in Tempo/examples/newUserGuide
 * (except twoTaskRace.tioa because it includes a simulation relation)
 */

/*
 * Top-level:
 * - removed 'simDef and simProg'
 * - simplified the invariant syntax
 */
spec : ( typeDecls | imports | funDecls | vocabDef | autoDef | invDef )+ ;

imports : 'imports' vocabRef (',' vocabRef )* ;

vocabRef : ID ( '(' actual (',' actual )* ')' )? ;

funDecls : 'let' (funDecl ';')+ ;

funDecl : ID '(' (ID (',' ID)* )? ')' ':' typeSignature '=' expr ;

invDef : 'invariant' idOrNumeral? 'of' ID ':' expr ';' ;

idOrNumerals : idOrNumeral ( ',' idOrNumeral )* ;

idOrNumeral : (ID | INT ) ;

actual : (expr | 'Type' typeRef ) ;


/*
 * Type declarations
 */
typeDecls : 'types' typeDecl ( ',' typeDecl )* ;

typeDecl : ID (':' typeRef)? ;

typeRef : ID |
          ID '[' typeList ']' |
          'Enumeration' '[' idOrNumerals ']' |
          'Tuple' '[' fieldDecls ']' |
          'Union' '[' fieldDecls ']' ;

fieldDecls : fieldDecl ( ',' fieldDecl )* ;

fieldDecl : ID ( ',' ID )* ':' typeRef ;

typeSignature : (typeList)? '->' typeRef ;

typeList : typeRef ( ',' typeRef )* ;


/*
 * Vocabulary definition: we don't support the infix/mixfix/prefix specs
 * operator declarations must be of the form ID: typeSignature
 */
vocabDef : 'vocabulary' ID (formals)? (importsAndTypes)? (operators)? 'end' ;

importsAndTypes : ( imports | typeDecls )+ ;

operators : 'operators' opDecl (',' opDecl )* ;

opDecl : ID ( ',' ID )* ':' typeSignature ;



/*
 * Automaton definitions
 * - no support for tasks and schedules
 */
autoDef : 'automaton' ID ( formals (where)? )? (imports)? autoCore ;

formals : '(' formal (',' formal )* ')' ;

formal : ID (',' ID)* ':' (typeRef | 'Type') ;

where : 'where' expr ;


autoCore : basicAutomaton | composedAutomaton ;

basicAutomaton : (actionSignature)? states (funDecls)? (transitions)? (trajectories)? ;


/*
 * Signature
 */
actionSignature : 'signature' (formalActions)+ ;

formalActions : 
      'input' formalAction ( ',' formalAction )* 
    | 'output' formalAction ( ',' formalAction )*
    | 'internal' formalAction ( ',' formalAction )* ;

formalAction : ID ( '(' sigFormal (',' sigFormal )* ')' (where)? )? ;

sigFormal : 'const' expr | ID (',' ID )* ':' typeRef ;



/*
 * State variable declaration + initialization
 * - note: initially seems to be an old form (could remove it?)
 */
states : 'states' (state ';')+ (initially)? | 'states' ;

state : ID ':' typeRef ( ':=' value )? ;

initially : 'initially' expr ';' ;




/*
 * Transitions: allowed only 'var := value' and if-then-else in the 'effStmt'
 * Also removed the 'ensuring' clause and 'urgency'
 *      simplified the precondition syntax to 'pre' expr ';'
 */
transitions : 'transitions' (transition)+ ;

transition :
      'input' transitionCore 
    | 'output' transitionCore 
    | 'internal' transitionCore ;

transitionCore : ID ( '(' actionActuals ')' (where)? )? (localVars)? (precondition)? (effect)? ;

actionActuals : expr (',' expr)* ;

localVars : 'locals' ( localDecl ';' )+ ;

precondition : 'pre' expr ';' ;

effect : 'eff' effProgram  ;

localDecl : ID ':' typeRef ( ':=' value ) ? ;

effProgram : ( effStmt )+ ;

effStmt : lvalue ':=' value 
    | 'if' effCondRec 'fi'
    | ';'
    ;

effCondRec : expr 'then' effProgram ( 'elseif' effCondRec | 'else' effProgram ) ? ;



/*
 * Trajectories:
 * - simplified the syntax: 
 *   Formals and local functions are not supported. 
 *   They don't seem to be use in any of the examples anyway.
 * - invariant clause: simplified to 'invariant expr ;'
 */
trajectories : 'trajectories' (trajectory)+ ;

trajectory : 'trajdef' ID (trajInvariant)? (stopCond)? (evolve)? ;

trajInvariant : 'invariant' expr ';' ;

stopCond : 'stop' 'when' expr ';' ;

evolve : 'evolve' ( evolveStmt)+  ;

evolveStmt : expr ';' ;



/*
 * Composed Automaton
 */
composedAutomaton : components (hiddenActionSets )? ;

components : 'components' ( component ';' )+ ;

component : ID ('[' varList (',' varList )* ']' )? (':' componentDef )? (where)? ;

componentDef : ID ( '(' actual (',' actual )* ')' )? ;

hiddenActionSets : 'hidden' ( actionSet ';' ) + ;

actionSet : ( compInstance '.')? ID ('(' expr (',' expr )* ')' )? ;

compInstance : ID ('[' expr (',' expr )* ']' )? ( '.' ( '[' expr (',' expr)* ']' )? )* ;

varList: ID (',' ID)* ':' typeRef ;


/*
 * Expressions
 */
expr : 
      'if' expr 'then' expr 'else' expr
    | expr BINARY_OP expr
    | UNARY_OP expr
    | QUANTIFIER ID ':' typeRef expr
    | '[' expr (',' expr)* ']'
    | '{' ID ':' typeRef 'where' expr '}'
    | '{' expr (',' expr)* '}'
    | expr ( '.' ID | '[' expr (',' expr)* ']' )+
    | ID '(' expr (',' expr)* ')'
    | '(' typeRef ')' expr
    | 'true'
    | 'false'
    | '\infty'
    | '{}'
    | INT
    | FLOAT
    | ID
    | ID '\''
    | '(' expr ')' ;



/*
 * Assignment: don't try to support the 'choice' construct
 */
lvalue : ID ( '[' expr (',' expr)* ']' | '.' ID )* ;

value : expr | choice ;

choice : 'choose' variable 'where' expr ;

variable : ID (':' typeRef)? ;


/*
 * TERMINALS
 */
ID : "[a-zA-Z][a-zA-Z_0-9\$]*" ;

INT : "[\+\-]?[0-9]+" ;

FLOAT : "[\+\-]?[0-9]+\.[0-9]+" 
  | "[\+\-]?[0-9]+(\.[0-9]+)? [Ee][\+\-][0-9]+" ;

QUANTIFIER : '\A' | '\E' ;

UNARY_OP : 
   '+'   $unary_op_right 40 |
   '-'   $unary_op_right 40 |
   '~'   $unary_op_right 40 ;

BINARY_OP :
   '**'        $binary_op_left 30 |
   '/'         $binary_op_left 29 |
   '*'         $binary_op_left 29 |
   '+'         $binary_op_left 28 |
   '-'         $binary_op_left 28 |
   '\I'        $binary_op_left 27 |
   '\U'        $binary_op_left 26 |
   '|-'        $binary_op_left 25 |
   '-|'        $binary_op_left 25 |
   '||'        $binary_op_left 24 |
   '<'         $binary_op_left 20 |
   '>'         $binary_op_left 20 |
   '<='        $binary_op_left 20 |
   '>='        $binary_op_left 20 |
   '\notin'    $binary_op_left 20 | 
   '\in'       $binary_op_left 20 |
   '\subset'   $binary_op_left 20 |
   '\subseteq' $binary_op_left 20 |
   '\supset'   $binary_op_left 20 |
   '\supseteq' $binary_op_left 20 |
   '='         $binary_op_left 19 |
   '~='        $binary_op_left 19 |
   '/\x5c'     $binary_op_left 6 |
   '\/'        $binary_op_left 5 |
   '=>'        $binary_op_left 4 |
   '<=>'       $binary_op_left 3 ;


whitespace : ( "[ \t\n\r\v\f]" | comment )* ;

comment :  '%' "[^\n]*" '\n'  ;
